import itertools
import pytest
from medimages4tests.dummy.dicom.mri.t1w.siemens.skyra.syngo_d13c import (
    get_image as get_dicom,
)
from medimages4tests.dummy.nifti import get_image as get_nifti
from fileformats.core.exceptions import FormatMismatchError
from fileformats.core import from_paths
from fileformats.medimage import DicomDir, DicomSeries


def test_dicom_identify():
    DicomDir(get_dicom())


def test_dicom_not_identify():
    with pytest.raises(FormatMismatchError, match="No directory paths provided"):
        DicomDir(get_nifti())


def test_series_from_paths(tmp_path):
    filesets = [
        DicomSeries.sample(tmp_path, seed=1),
        DicomSeries.sample(tmp_path, seed=2),
        DicomSeries.sample(tmp_path, seed=3),
    ]

    fspaths = list(itertools.chain(*(f.fspaths for f in filesets)))

    detected = from_paths(fspaths, DicomSeries)

    assert set(detected) == set(filesets)


def test_dicom_series_metadata(tmp_path):
    series = DicomSeries.sample(tmp_path)

    # Check series number is not a list
    assert not isinstance(series.metadata["SeriesNumber"], list)
    # check the SOP Instance ID has been converted into a list
    assert isinstance(series.metadata["SOPInstanceUID"], list)


def test_dicom_dir_metadata(tmp_path):
    series = DicomDir.sample(tmp_path)

    # Check series number is not a list
    assert not isinstance(series.metadata["SeriesNumber"], list)
    # check the SOP Instance ID has been converted into a list
    assert isinstance(series.metadata["SOPInstanceUID"], list)



import typing as ty
from pathlib import Path
import tempfile
from fileformats.core import extra_implementation, FileSet
from fileformats.medimage import MedicalImage


@extra_implementation(MedicalImage.deidentify)
def no_deidentification_necessary(
    image: MedicalImage,
    out_dir: ty.Optional[Path] = None,
    new_stem: ty.Optional[str] = None,
    copy_mode: FileSet.CopyMode = FileSet.CopyMode.copy,
) -> MedicalImage:
    """Assume that no deidentification is needed for medical images by default. We make a
    copy of the image in the output directory for consistency with the behavior of other
    deidentification formats"""
    if image.contains_phi:
        raise NotImplementedError(
            f"{type(image)} images contain Protected Health Information (PHI) and needs a "
            "specific deidentification method"
        )
    if out_dir is None:
        out_dir = Path(tempfile.mkdtemp())
    out_dir.mkdir(exist_ok=True, parents=True)
    cpy = image.copy(out_dir, new_stem=new_stem, mode=copy_mode)
    return cpy



from . import MrProcedure


class TissueContrast(MrProcedure):
    pass


class T1Weighted(TissueContrast):
    ontology_link = "http://www.radlex.org/RID/RID10794"
    description = "T1-weighted MRI image contrast"


class T2Weighted(TissueContrast):
    ontology_link = "http://www.radlex.org/RID/RID10795"
    description = "T2-weighted MRI image contrast"


class T2StarWeighted(TissueContrast):
    ontology_link = "http://www.radlex.org/RID/RID10796"
    description = "T2*-weighted MRI image contrast"


class T1T2Weighted(TissueContrast):
    ontology_link = "http://www.radlex.org/RID/RID10796"
    description = "T1/T2-weighted MRI image contrast"


class DiffusionWeighted(TissueContrast):
    ontology_link = "http://www.radlex.org/RID/RID10799"
    description = (
        "Assessment of the random Brownian motion of water molecules within the voxel."
    )


class FluidAttenuatedInversionRecovery(TissueContrast):
    ontology_link = "http://www.radlex.org/RID/RID35806"
    description = "Fluid Attenuated Inversion Recovery (FLAIR)"


class IntermediateWeighted(TissueContrast):
    ontology_link = "http://www.radlex.org/RID/RID10792"
    description = "Intermediate (proton density) weighted MRI image contrast"


Dwi = DiffusionWeighted
T1w = T1Weighted
T2w = T2Weighted
T2sw = T2StarWeighted
Flair = FluidAttenuatedInversionRecovery
T1T2w = T1T2Weighted

ProtonDensity = IntermediateWeighted



import numpy as np

from fileformats.medimage import (
    NiftiGzX,
    NiftiGzXBvec,
    T1w,
    Dmri,
    Fmri,
    Bval,
    Bvec,
)


def test_t1w_generator():
    img = NiftiGzX[T1w].sample()
    assert len(img.dims()) == 3


def test_fmri_generator():
    img = NiftiGzX[Fmri].sample()
    assert len(img.dims()) == 4


def test_dmri_generator():
    img = NiftiGzXBvec[Dmri].sample()
    assert len(img.dims()) == 4


def test_bval_generator():
    bval = Bval.sample()
    assert len(bval.read_array()) >= 5


def test_bvec_generator():
    bvec = Bvec.sample()
    n, m = bvec.encodings_array.shape
    assert n >= 5
    assert m == 4



import pytest
from fileformats.medimage import DicomImage, DicomDir, DicomSeries, Nifti1
from medimages4tests.dummy.dicom.mri.t1w.siemens.skyra.syngo_d13c import (
    get_image as get_dicom_image,
)


@pytest.fixture(params=["image", "dir", "series"])
def dicom(request):
    dicom_dir = get_dicom_image(first_name="John", last_name="Doe")
    dicom_files = (p for p in dicom_dir.iterdir() if p.suffix == ".dcm")
    if request.param == "image":
        return DicomImage(next(dicom_files))
    elif request.param == "dir":
        return DicomDir(dicom_dir)
    else:
        return DicomSeries(dicom_files)


def test_deidentify_dicom(dicom):
    assert str(dicom.metadata["PatientName"]) == "Doe^John"
    assert dicom.metadata["InstitutionAddress"]
    assert not dicom.metadata["PatientBirthDate"].endswith("0101")
    deidentified = dicom.deidentify()
    assert str(deidentified.metadata["PatientName"]) == "Anonymous^Anonymous"
    assert deidentified.metadata["InstitutionAddress"] == ""
    assert deidentified.metadata["PatientBirthDate"] == "19800101"


def test_nifti_deidentify():
    nifti = Nifti1.sample()
    deidentified = nifti.deidentify()
    assert nifti is not deidentified
    assert nifti.hash_files() == deidentified.hash_files()



from ._version import __version__
from .base import MedicalImagingData, MedicalImage


from .misc import (
    Analyze,
    Mgh,
    MghGz,
)
from .nifti import (
    Nifti,
    Nifti1,
    Nifti2,
    NiftiGz,
    NiftiX,
    NiftiGzX,
)
from .diffusion import (
    DwiEncoding,
    Bvec,
    Bval,
    NiftiBvec,
    NiftiGzBvec,
    NiftiXBvec,
    NiftiGzXBvec,
)
from .dicom import (
    DicomImage,
    DicomCollection,
    DicomDir,
    DicomSeries,
    # Vnd_Siemens_Vision,
    # Vnd_Siemens_VisionDir,
)
from .raw import (
    Kspace,
    Rda,
    PetRawData,
    PetListMode,
    PetSinogram,
    PetCountRate,
    PetNormalisation,
)
from .surface import Gifti
from .contents.imaging.modality import (
    ImagingModality,
    CombinedModalities,
    DualEnergyXrayAbsorptiometry,
    Fluoroscopy,
    MrFluoroscopy,
    RadioFluoroscopy,
    MagneticResonanceImaging,
    DiffusionTensorImaging,
    DynamicContrast,
    EnhancedMagneticResonanceImaging,
    FunctionalMagneticResonanceImaging,
    MagneticResonanceAngiography,
    MagneticResonanceSpectroscopy,
    NuclearMedicineImaging,
    PositronEmissionTomography,
    PanographicRadiograph,
    ProjectionRadiography,
    ComputedRadiography,
    DigitalRadiography,
    DualEnergySubtractionRadiograpgy,
    Mammography,
    ScreenFilmRadiography,
    Stereoscopy,
    StereotacticRadiography,
    Spectroscopy,
    Tomography,
    ComputedTomography,
    Ultrasound,
    Mri,
    Pet,
    Dti,
    Dmri,
    Fmri,
    Cr,
    Ct,
    Dx,
    Mg,
    Mr,
    Nm,
    Pt,
    Px,
    Rf,
    Rg,
    Us,
)
from .contents.imaging.derivatives import Derivative, Mask
from .contents.anatomical_entity.material_anatomical_entity.anatomical_structure import (
    Brain,
    SpinalCord,
)
from .contents.property.imaging_procedure.cross_sectional_procedure.mr_procedure.tissue_contrast import (
    T1w,
    T2w,
    T2sw,
    T1T2w,
    Flair,
    Dwi,
    T2Weighted,
    T1Weighted,
    T2StarWeighted,
    T1T2Weighted,
    DiffusionWeighted,
    FluidAttenuatedInversionRecovery,
    IntermediateWeighted,
)
from .itk import (
    GDCM,
    GIPL,
    VTK,
    PGM,
    MetaImage,
    Nrrd,
    NrrdGz,
    ItkImage,
    ItkAll,
)


__all__ = [
    "__version__",
    "MedicalImagingData",
    "MedicalImage",
    "DicomImage",
    "Analyze",
    "Mgh",
    "MghGz",
    "Nifti",
    "Nifti1",
    "Nifti2",
    "NiftiGz",
    "NiftiX",
    "NiftiGzX",
    "DwiEncoding",
    "Bvec",
    "Bval",
    "NiftiBvec",
    "NiftiGzBvec",
    "NiftiXBvec",
    "NiftiGzXBvec",
    "DicomCollection",
    "DicomDir",
    "DicomSeries",
    "Kspace",
    "Rda",
    "PetRawData",
    "PetListMode",
    "PetSinogram",
    "PetCountRate",
    "PetNormalisation",
    "Gifti",
    "ImagingModality",
    "CombinedModalities",
    "DualEnergyXrayAbsorptiometry",
    "Fluoroscopy",
    "MrFluoroscopy",
    "RadioFluoroscopy",
    "MagneticResonanceImaging",
    "DiffusionTensorImaging",
    "DynamicContrast",
    "EnhancedMagneticResonanceImaging",
    "FunctionalMagneticResonanceImaging",
    "MagneticResonanceAngiography",
    "MagneticResonanceSpectroscopy",
    "NuclearMedicineImaging",
    "PositronEmissionTomography",
    "PanographicRadiograph",
    "ProjectionRadiography",
    "ComputedRadiography",
    "DigitalRadiography",
    "DualEnergySubtractionRadiograpgy",
    "Mammography",
    "ScreenFilmRadiography",
    "Stereoscopy",
    "StereotacticRadiography",
    "Spectroscopy",
    "Tomography",
    "ComputedTomography",
    "Ultrasound",
    "Mri",
    "Pet",
    "Dti",
    "Dmri",
    "Fmri",
    "Cr",
    "Ct",
    "Dx",
    "Mg",
    "Mr",
    "Nm",
    "Pt",
    "Px",
    "Rf",
    "Rg",
    "Us",
    "Derivative",
    "Mask",
    "Brain",
    "SpinalCord",
    "T1w",
    "T2w",
    "T2sw",
    "T1T2w",
    "Flair",
    "Dwi",
    "T2Weighted",
    "T1Weighted",
    "T2StarWeighted",
    "T1T2Weighted",
    "DiffusionWeighted",
    "FluidAttenuatedInversionRecovery",
    "IntermediateWeighted",
    "GDCM",
    "GIPL",
    "VTK",
    "PGM",
    "MetaImage",
    "Nrrd",
    "NrrdGz",
    "ItkImage",
    "ItkAll",
]



import sys
import typing as ty
from pathlib import Path
from fileformats.core import FileSet, extra
from fileformats.medimage import MedicalImagingData
from fileformats.generic import BinaryFile

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class PetRawData(BinaryFile, MedicalImagingData):
    """Base class for raw PET data files"""

    @extra
    def deidentify(
        self,
        out_dir: ty.Optional[Path] = None,
        new_stem: ty.Optional[str] = None,
        copy_mode: FileSet.CopyMode = FileSet.CopyMode.copy,
    ) -> Self:
        """Returns a new copy of the data with any subject-identifying information
        stripped from the from the data header"""
        raise NotImplementedError


class PetListMode(PetRawData):
    "raw projection data"


class PetSinogram(PetRawData):
    "histogrammed projection data in a reconstruction-friendly format"


class PetCountRate(PetRawData):
    "number of prompt/random/single events per unit time"


class PetNormalisation(PetRawData):
    "normalisation scan or the current cross calibration factor"


class PetPhysio(PetRawData):
    "physiological data (e.g. ECG, respiration) for motion correction"


class PetParameterisation(PetRawData):
    "physiological data (e.g. ECG, respiration) for motion correction"



import typing
from fileformats.core import extra, validated_property
from fileformats.core.typing import TypeAlias
from fileformats.core.mixin import WithAdjacentFiles
from fileformats.generic import BinaryFile
from .nifti import NiftiGzX, NiftiGz, Nifti1, NiftiX


if typing.TYPE_CHECKING:
    import numpy.typing  # noqa: F401

EncodingArrayType: TypeAlias = (
    typing.Any
)  # In Py<3.9 this is problematic "numpy.typing.NDArray[numpy.floating[typing.Any]]"


class DwiEncoding:
    @extra
    def read_encodings(self) -> EncodingArrayType:
        "Both the gradient direction and weighting combined into a single Nx4 array"
        raise NotImplementedError

    @property
    def encodings_array(self) -> EncodingArrayType:
        return self.read_encodings()

    @property
    def directions(self) -> EncodingArrayType:
        "gradient direction and weighting combined into a single Nx4 array"
        return self.encodings_array[:, :3]

    @property
    def b_values(self) -> EncodingArrayType:
        "the b-value weighting"
        return self.encodings_array[:, 3]


class Bval(BinaryFile):

    ext = ".bval"

    @extra
    def read_array(self) -> EncodingArrayType:
        raise NotImplementedError


class Bvec(WithAdjacentFiles, DwiEncoding, BinaryFile):
    """FSL-style diffusion encoding, in two separate files"""

    ext = ".bvec"

    @validated_property
    def b_values_file(self) -> Bval:
        return Bval(self.select_by_ext(Bval))

    @validated_property
    def num_encodings(self) -> int:
        num_bvals = len(self.b_values_file.read_array())
        num_bvecs = len(self.directions)
        if num_bvals != num_bvecs:
            raise ValueError(
                f"The number of b-values ({num_bvals}) does not match the number of "
                f"directions {num_bvecs}"
            )
        return num_bvals


# NIfTI file format gzipped with BIDS side car
class WithBvec(WithAdjacentFiles):
    @validated_property
    def encoding(self) -> Bvec:
        return Bvec(self.select_by_ext(Bvec))  # type: ignore[attr-defined]


class NiftiBvec(WithBvec, Nifti1):
    iana_mime = "application/x-nifti2+bvec"


class NiftiGzBvec(WithBvec, NiftiGz):
    iana_mime = "application/x-nifti2+gzip.bvec"


class NiftiXBvec(WithBvec, NiftiX):
    iana_mime = "application/x-nifti2+json.bvec"


class NiftiGzXBvec(WithBvec, NiftiGzX):
    iana_mime = "application/x-nifti2+gzip.json.bvec"



from fileformats.generic import BinaryFile


class Kspace(BinaryFile):

    binary = True
    # iana_mime = None
    pass


class Rda(BinaryFile):
    """MRS format"""

    ext = ".rda"
    binary = True



from fileformats.generic import BinaryFile
from fileformats.core import validated_property
from fileformats.core.mixin import WithSideCars, WithMagicNumber, WithAdjacentFiles
from fileformats.application import Json
from fileformats.application.archive import BaseGzip
from .base import MedicalImage


class Nifti(MedicalImage, BinaryFile):

    ext: str = ".nii"
    contains_phi = False


class WithBids(WithSideCars):

    primary_type = Nifti
    side_car_types = (Json,)

    @validated_property
    def json_file(self) -> Json:
        return Json(self.select_by_ext(Json))  # type: ignore[attr-defined]


class Nifti1(WithMagicNumber, Nifti):

    iana_mime = "application/x-nifti1"
    magic_number = "6E2B3100"
    magic_number_offset = 344


class Nifti2(WithMagicNumber, Nifti):

    iana_mime = "application/x-nifti2"
    magic_number = "6e2b3200"
    magic_number_offset = 344


class NiftiGz(Nifti, BaseGzip):
    ext = ".nii.gz"
    iana_mime = "application/x-nifti1+gzip"
    archived_type = Nifti


class NiftiX(WithBids, Nifti):
    iana_mime = "application/x-nifti1+json"


class NiftiGzX(WithBids, NiftiGz):
    iana_mime = "application/x-nifti1+gzip.bids"


class NiftiDataFile(MedicalImage):

    ext = ".img"


class NiftiWithDataFile(WithAdjacentFiles, Nifti1):
    """Nifti file with separate data file"""

    magic_number = "6E693100"
    alternate_exts = (".hdr",)

    @validated_property
    def data_file(self) -> NiftiDataFile:
        return NiftiDataFile(self.select_by_ext(NiftiDataFile))



from .base import (
    PetRawData,
    PetListMode,
    PetSinogram,
    PetCountRate,
    PetNormalisation,
)

__all__ = [
    "PetRawData",
    "PetListMode",
    "PetSinogram",
    "PetCountRate",
    "PetNormalisation",
]



from pathlib import Path
import typing as ty
import nibabel
import typing  # noqa: F401
import numpy.typing  # noqa: F401
from fileformats.core import FileSet, SampleFileGenerator, extra_implementation
from fileformats.medimage import (
    MedicalImage,
    Nifti,
    NiftiGz,
    Nifti1,
    NiftiGzX,
    NiftiGzXBvec,
    NiftiXBvec,
    NiftiX,
    T1w,
    Fmri,
    Dmri,
    Brain,
)
from fileformats.medimage.base import DataArrayType
import medimages4tests.dummy.nifti
import medimages4tests.mri.neuro.t1w
import medimages4tests.mri.neuro.dwi
import medimages4tests.mri.neuro.bold


@extra_implementation(FileSet.read_metadata)
def nifti_read_metadata(nifti: Nifti, **kwargs: ty.Any) -> ty.Mapping[str, ty.Any]:
    metadata = dict(nibabel.load(nifti.fspath).header)  # type: ignore[call-overload, attr-defined]
    return metadata  # type: ignore[no-any-return]


@extra_implementation(MedicalImage.read_array)
def nifti_data_array(nifti: Nifti) -> DataArrayType:  # noqa
    return nibabel.load(nifti.fspath).get_data()  # type: ignore[attr-defined]


@extra_implementation(MedicalImage.vox_sizes)
def nifti_vox_sizes(nifti: Nifti) -> ty.Tuple[float, float, float]:
    ndims = len(nifti_dims(nifti))
    return tuple(float(d) for d in nifti.metadata["pixdim"][1 : ndims + 1])  # type: ignore[return-value]


@extra_implementation(MedicalImage.dims)
def nifti_dims(nifti: Nifti) -> ty.Tuple[int, int, int]:
    dim_array = [int(d) for d in nifti.metadata["dim"]]
    for i in range(1, len(dim_array)):
        if all(d == 1 for d in dim_array[i:]):
            break  # Stop when the remaining dimensions are singletons
    return tuple(dim_array[1:i])  # type: ignore[return-value]


@extra_implementation(FileSet.generate_sample_data)
def nifti_generate_sample_data(
    nifti: Nifti1,
    generator: SampleFileGenerator,
) -> ty.List[Path]:
    return [
        medimages4tests.dummy.nifti.get_image(
            out_file=generator.generate_fspath(file_type=Nifti1),
            seed=generator.seed,
        )
    ]


@extra_implementation(FileSet.generate_sample_data)
def nifti_gz_generate_sample_data(
    nifti: NiftiGz,
    generator: SampleFileGenerator,
) -> ty.List[Path]:
    return [
        medimages4tests.dummy.nifti.get_image(
            out_file=generator.generate_fspath(file_type=NiftiGz),
            compressed=True,
            seed=generator.seed,
        )
    ]


@extra_implementation(FileSet.generate_sample_data)
def t1w_nifti_gz_x_generate_sample_data(
    nifti: NiftiGzX[T1w],  # type: ignore[type-arg]
    generator: SampleFileGenerator,
) -> ty.List[Path]:
    return _get_t1w_nifti_gz_x(generator)


@extra_implementation(FileSet.generate_sample_data)
def t1w_nifti_x_generate_sample_data(
    nifti: NiftiX[T1w],  # type: ignore[type-arg]
    generator: SampleFileGenerator,
) -> ty.List[Path]:
    nifti_gz_x = NiftiGzX(_get_t1w_nifti_gz_x(generator))
    return NiftiX.convert(nifti_gz_x)  # type: ignore[return-value]


@extra_implementation(FileSet.generate_sample_data)
def t1w_brain_nifti_gz_x_generate_sample_data(
    nifti: NiftiGzX[T1w, Brain],  # type: ignore[type-arg]
    generator: SampleFileGenerator,
) -> ty.List[Path]:
    return _get_t1w_nifti_gz_x(generator)


@extra_implementation(FileSet.generate_sample_data)
def t1w_brain_nifti_x_generate_sample_data(
    nifti: NiftiX[T1w, Brain],  # type: ignore[type-arg]
    generator: SampleFileGenerator,
) -> ty.List[Path]:
    nifti_gz_x = NiftiGzX(_get_t1w_nifti_gz_x(generator))
    return NiftiX.convert(nifti_gz_x)  # type: ignore[return-value]


@extra_implementation(FileSet.generate_sample_data)
def fmri_brain_nifti_gz_x_generate_sample_data(
    nifti: NiftiGzX[Fmri, Brain],  # type: ignore[type-arg]
    generator: SampleFileGenerator,
) -> ty.List[Path]:
    return _get_fmri_nifti_gz_x(generator)


@extra_implementation(FileSet.generate_sample_data)
def fmri_brain_nifti_x_generate_sample_data(
    nifti: NiftiX[Fmri, Brain],  # type: ignore[type-arg]
    generator: SampleFileGenerator,
) -> ty.List[Path]:
    nifti_gz_x = NiftiGzX(_get_fmri_nifti_gz_x(generator))
    return NiftiX.convert(nifti_gz_x)  # type: ignore[return-value]


@extra_implementation(FileSet.generate_sample_data)
def fmri_nifti_gz_x_generate_sample_data(
    nifti: NiftiGzX[Fmri],  # type: ignore[type-arg]
    generator: SampleFileGenerator,
) -> ty.List[Path]:
    return _get_fmri_nifti_gz_x(generator)


@extra_implementation(FileSet.generate_sample_data)
def fmri_nifti_x_generate_sample_data(
    nifti: NiftiX[Fmri],  # type: ignore[type-arg]
    generator: SampleFileGenerator,
) -> ty.List[Path]:
    nifti_gz_x = NiftiGzX(_get_fmri_nifti_gz_x(generator))
    return NiftiX.convert(nifti_gz_x)  # type: ignore[return-value]


@extra_implementation(FileSet.generate_sample_data)
def dmri_brain_nifti_gz_x_generate_sample_data(
    nifti: NiftiGzXBvec[Dmri, Brain],  # type: ignore[type-arg]
    generator: SampleFileGenerator,
) -> ty.List[Path]:
    return _get_dmri_nifti_gz_x(generator)


@extra_implementation(FileSet.generate_sample_data)
def dmri_brain_nifti_x_generate_sample_data(
    nifti: NiftiXBvec[Dmri, Brain],  # type: ignore[type-arg]
    generator: SampleFileGenerator,
) -> ty.List[Path]:
    nifti_gz_x = NiftiGzX(_get_dmri_nifti_gz_x(generator))
    return NiftiX.convert(nifti_gz_x)  # type: ignore[return-value]


@extra_implementation(FileSet.generate_sample_data)
def dmri_nifti_gz_x_generate_sample_data(
    nifti: NiftiGzXBvec[Dmri],  # type: ignore[type-arg]
    generator: SampleFileGenerator,
) -> ty.List[Path]:
    return _get_dmri_nifti_gz_x(generator)


@extra_implementation(FileSet.generate_sample_data)
def dmri_nifti_x_generate_sample_data(
    nifti: NiftiXBvec[Dmri],  # type: ignore[type-arg]
    generator: SampleFileGenerator,
) -> ty.List[Path]:
    nifti_gz_x = NiftiGzXBvec(_get_dmri_nifti_gz_x(generator))
    return NiftiX.convert(nifti_gz_x)  # type: ignore[return-value]


def _get_t1w_nifti_gz_x(generator: SampleFileGenerator) -> ty.List[Path]:
    sample = generator.seed if generator.seed else "ds002014-01"
    fspaths = medimages4tests.mri.neuro.t1w.get_image(sample=sample)
    NiftiGzX(fspaths).copy(generator.dest_dir, mode=NiftiGzX.CopyMode.link_or_copy)
    return fspaths  # type: ignore[no-any-return]


def _get_fmri_nifti_gz_x(generator: SampleFileGenerator) -> ty.List[Path]:
    sample = generator.seed if generator.seed else "ds002014-01"
    fspaths = medimages4tests.mri.neuro.bold.get_image(sample=sample)
    NiftiGzX(fspaths).copy(generator.dest_dir, mode=NiftiGzX.CopyMode.link_or_copy)
    return fspaths  # type: ignore[no-any-return]


def _get_dmri_nifti_gz_x(generator: SampleFileGenerator) -> ty.List[Path]:
    sample = generator.seed if generator.seed else "ds004024-CON031"
    fspaths = medimages4tests.mri.neuro.dwi.get_image(sample=sample)
    NiftiGzX(fspaths).copy(generator.dest_dir, mode=NiftiGzX.CopyMode.link_or_copy)
    return fspaths  # type: ignore[no-any-return]



import typing as ty
from .. import ContentsClassifier


class ImagingModality(ContentsClassifier):
    ontology_link = "http://www.radlex.org/RID/RID10311"
    description: ty.Optional[str] = (
        "Form of imaging that depends on the way the image is produced"
    )
    dicom_modality: ty.Optional[str] = None


class CombinedModalities(ImagingModality):
    ontology_link = "http://www.radlex.org/RID/RID49580"
    description = (
        "These are cases where 2 different modalities are "
        "performed in the same imaging setup without moving the "
        "patient. These classes were created as sets, with the "
        "individual modalities as the set members. Any reasoning "
        "involving modality should accommodate this."
    )


class DualEnergyXrayAbsorptiometry(ImagingModality):
    ontology_link = "http://www.radlex.org/RID/RID10363"
    description = None


class Fluoroscopy(ImagingModality):
    ontology_link = "http://www.radlex.org/RID/RID10361"
    description: ty.Optional[str] = None


class MrFluoroscopy(Fluoroscopy):
    ontology_link = "http://www.radlex.org/RID/RID10319"
    description = (
        "Non-invasive method of vascular imaging and determination "
        "of internal anatomy without injection of contrast media or "
        "radiation exposure. The technique is used especially in "
        "cerebral angiography as well as for studies of other vascular "
        "structures. [MeSH]"
    )


class RadioFluoroscopy(Fluoroscopy):
    ontology_link = "http://www.radlex.org/RID/RID45709"
    description = (
        "Production of an image when x-rays strike a fluorescent screen. [MeSH]"
    )


class MagneticResonanceImaging(ImagingModality):
    ontology_link = "http://www.radlex.org/RID/RID10312"
    description: ty.Optional[str] = (
        "Non-invasive method of demonstrating internal anatomy "
        "based on the principle that atomic nuclei in a strong "
        "magnetic field absorb pulses of radiofrequency energy and "
        "emit them as radiowaves which can be reconstructed into "
        "computerized images. The concept includes proton spin "
        "tomographic techniques. [MeSH]"
    )
    dicom_modality = "MR"


class DiffusionMagneticResonanceImaging(MagneticResonanceImaging):
    ontology_link = "http://www.radlex.org/RID/RID38778"
    description = None


class DiffusionTensorImaging(DiffusionMagneticResonanceImaging):
    ontology_link = "http://www.radlex.org/RID/RID38778"
    description = None


class DynamicContrast(MagneticResonanceImaging):
    ontology_link = "http://www.radlex.org/RID/RID49531"
    description = (
        "An imaging method with a timed series of T1-weighted "
        "images used to detect and measure signal intensity "
        "change (enhancement) over time following administration "
        "of intravenous contrast agent to noninvasively access "
        "tissue vascular characteristics."
    )


class EnhancedMagneticResonanceImaging(MagneticResonanceImaging):
    pass


class FunctionalMagneticResonanceImaging(MagneticResonanceImaging):
    ontology_link = "http://www.radlex.org/RID/RID10317"
    description = None


class MagneticResonanceAngiography(MagneticResonanceImaging):
    ontology_link = "http://www.radlex.org/RID/RID10319"
    description = None


class MagneticResonanceSpectroscopy(ImagingModality):
    ontology_link = "http://www.radlex.org/RID/RID10315"
    description = (
        "	Spectroscopic method of measuring the magnetic moment of "
        "elementary particles such as atomic nuclei, protons or "
        "electrons. It is employed in clinical applications such as "
        "nmr tomography (magnetic resonance imaging). [MeSH]"
    )


class NuclearMedicineImaging(ImagingModality):
    ontology_link = "http://www.radlex.org/RID/RID10330"
    description: ty.Optional[str] = None
    dicom_modality = "NM"


class PositronEmissionTomography(NuclearMedicineImaging):
    ontology_link = "http://www.radlex.org/RID/RID10337"
    description = (
        "An imaging technique using compounds labelled with "
        "short-lived positron-emitting radionuclides (such as "
        "carbon-11, nitrogen-13, oxygen-15 and fluorine-18) to "
        "measure cell metabolism. It has been useful in study of "
        "soft tissues such as cancer; cardiovascular system; and "
        "brain. SPECT is closely related to PET, but uses isotopes "
        "with longer half-lives and resolution is lower. [MeSH]"
    )
    dicom_modality = "PT"


class PanographicRadiograph(ImagingModality):
    ontology_link = "http://www.radlex.org/RID/RID10360"
    description = None
    dicom_modality = "PX"


class ProjectionRadiography(ImagingModality):
    ontology_link = "http://www.radlex.org/RID/RID10345"
    description: ty.Optional[str] = (
        "Examination of any part of the body for diagnostic purposes "
        "by means of roentgen rays, recording the image on a "
        "sensitized surface (such as photographic film). [MeSH]"
    )


class ComputedRadiography(ProjectionRadiography):
    ontology_link = "http://www.radlex.org/RID/RID10349"
    dicom_modality = "CR"
    description = None


class DigitalRadiography(ProjectionRadiography):
    ontology_link = "http://www.radlex.org/RID/RID10351"
    dicom_modality = "DR"
    description = None


class DualEnergySubtractionRadiograpgy(ProjectionRadiography):
    ontology_link = "http://www.radlex.org/RID/RID10356"
    description = None


class Mammography(ProjectionRadiography):
    ontology_link = "http://www.radlex.org/RID/RID10357"
    description = None


class ScreenFilmRadiography(ProjectionRadiography):
    ontology_link = "http://www.radlex.org/RID/RID10353"
    description = "Conventional radiography"
    dicom_modality = "RG"


class Stereoscopy(ProjectionRadiography):
    ontology_link = "http://www.radlex.org/RID/RID50131"
    description = None


class StereotacticRadiography(ProjectionRadiography):
    ontology_link = "http://www.radlex.org/RID/RID50260"
    description = None


class Spectroscopy(ImagingModality):
    ontology_link = "http://www.radlex.org/RID/RID10377"
    description = (
        "The measurement of the amplitude of the components of a "
        "complex waveform throughout the frequency range of the "
        "waveform. (McGraw-Hill Dictionary of Scientific and "
        "Technical Terms, 4th ed) [MeSH]"
    )


class Tomography(ImagingModality):
    ontology_link = "http://www.radlex.org/RID/RID28840"
    description: ty.Optional[str] = None


class ComputedTomography(Tomography):
    ontology_link = "http://www.radlex.org/RID/RID10321"
    description = (
        "Tomography using x-ray transmission and a computer algorithm "
        "to reconstruct the image. [MeSH]"
    )
    dicom_modality = "CT"


class Ultrasound(ImagingModality):
    ontology_link = "http://www.radlex.org/RID/RID10326"
    description = (
        "The visualization of deep structures of the body by "
        "recording the reflections of echoes of pulses of ultrasonic "
        "waves directed into the tissues. Use of ultrasound for "
        "imaging or diagnostic purposes employs frequencies ranging "
        "from 1.6 to 10 megahertz. [MeSH]"
    )
    dicom_modality = "US"


# Some extra abbreviations
Mri = MagneticResonanceImaging
Pet = PositronEmissionTomography
Dmri = DiffusionMagneticResonanceImaging
Dti = DiffusionTensorImaging
Fmri = FunctionalMagneticResonanceImaging

# DICOM abbreviations, taken from: https://dicom.nema.org/medical/dicom/current/output/chtml/part16/sect_CID_29.html
Cr = ComputedRadiography
Ct = ComputedTomography
Dx = DigitalRadiography
Mg = Mammography
Mr = MagneticResonanceImaging
Nm = NuclearMedicineImaging
Pt = PositronEmissionTomography
Px = PanographicRadiograph
Rf = RadioFluoroscopy
Rg = ScreenFilmRadiography
Us = Ultrasound
# AR = Autorefraction
# BI = Biomagnetic Imaging
# BMD = Bone Mineral Densitometry
# EPS = Cardiac Electrophysiology
# DMS = Dermoscopy
# DG = Diaphanography
# ECG = Electrocardiography
# EEG = Electroencephalography
# EMG = Electromyography
# EOG = Electrooculography
# ES = Endoscopy
# XC = External-camera Photography
# GM = General Microscopy
# HD = Hemodynamic Waveform
# IO = Intra-oral Radiography
# IVOCT = Intravascular Optical Coherence Tomography
# IVUS = Intravascular Ultrasound
# KER = Keratometry
# LS = Laser Scan
# LEN = Lensometry
# OAM = Ophthalmic Axial Measurements
# OPM = Ophthalmic Mapping
# OP = Ophthalmic Photography
# OPT = Ophthalmic Tomography
# OPTBSV = Ophthalmic Tomography B-scan Volume Analysis
# OPTENF = Ophthalmic Tomography En Face
# OPV = Ophthalmic Visual Field
# OCT = Optical Coherence Tomography
# OSS = Optical Surface Scanner
# PA = Photoacoustic
# POS = Position Sensor
# RESP = Respiratory Waveform
# RTIMAGE = RT Image
# SM = Slide Microscopy
# SRF = Subjective Refraction
# TG = Thermography
# BDUS = Ultrasound Bone Densitometry
# VA = Visual Acuity
# XA = X-Ray Angiography



from fileformats.core import from_mime
from fileformats.medimage import (
    NiftiGz,
    NiftiGzX,
    T1w,
)


def test_image_contents1():
    assert from_mime("medimage/t1w+nifti-gz") == NiftiGz[T1w]


def test_image_contents_generation():
    img = NiftiGzX[T1w].sample()
    assert len(img.dims()) == 3



from pathlib import Path
import typing as ty
import tempfile
import pydicom
import numpy
import numpy.typing
from fileformats.core import FileSet, extra_implementation
from fileformats.core import SampleFileGenerator
from fileformats.medimage import (
    MedicalImage,
    DicomImage,
    DicomCollection,
    DicomDir,
    DicomSeries,
)
import fileformats.extras.application.medical  # noqa: F401
from fileformats.medimage.base import DataArrayType
import medimages4tests.dummy.dicom.mri.t1w.siemens.skyra.syngo_d13c


@extra_implementation(MedicalImage.read_array)
def dicom_read_array(
    collection: DicomCollection,
) -> DataArrayType:
    image_stack = []
    for dcm_file in collection.contents:
        image_stack.append(pydicom.dcmread(dcm_file).pixel_array)
    return numpy.asarray(image_stack)


@extra_implementation(MedicalImage.vox_sizes)
def dicom_vox_sizes(collection: DicomCollection) -> ty.Tuple[float, float, float]:
    return tuple(
        collection.metadata["PixelSpacing"] + [collection.metadata["SliceThickness"]]
    )


@extra_implementation(MedicalImage.dims)
def dicom_dims(collection: DicomCollection) -> ty.Tuple[int, int, int]:
    return tuple(
        (
            collection.metadata["Rows"],
            collection.metadata["DataColumns"],
            len(list(collection.contents)),
        ),
    )


@extra_implementation(DicomCollection.series_number)
def dicom_series_number(collection: DicomCollection) -> str:
    return str(collection.metadata["SeriesNumber"])


@extra_implementation(FileSet.generate_sample_data)
def dicom_dir_generate_sample_data(
    dcmdir: DicomDir,
    generator: SampleFileGenerator,
) -> ty.List[Path]:
    dcm_dir = medimages4tests.dummy.dicom.mri.t1w.siemens.skyra.syngo_d13c.get_image()
    series_number = generator.rng.randint(1, SERIES_NUMBER_RANGE)
    dest = generator.generate_fspath(DicomDir)
    dest.mkdir()
    for dcm_file in dcm_dir.iterdir():
        dcm = pydicom.dcmread(dcm_file)
        dcm.SeriesNumber = series_number
        pydicom.dcmwrite(dest / dcm_file.name, dcm)
    return [dest]


@extra_implementation(FileSet.generate_sample_data)
def dicom_series_generate_sample_data(
    dcm_series: DicomSeries,
    generator: SampleFileGenerator,
) -> ty.List[Path]:
    dicom_dir: Path = dicom_dir_generate_sample_data(dcm_series, generator=generator)[0]  # type: ignore[arg-type]
    stem = generator.generate_fspath().stem
    fspaths = []
    for i, dicom_file in enumerate(dicom_dir.iterdir(), start=1):
        fspaths.append(dicom_file.rename(generator.dest_dir / f"{stem}-{i}.dcm"))
    dicom_dir.rmdir()
    return fspaths


SERIES_NUMBER_TAG = ("0020", "0011")
SERIES_NUMBER_RANGE = int(1e8)


@extra_implementation(MedicalImage.deidentify)
def dicom_deidentify(
    dicom: DicomImage,
    out_dir: ty.Optional[Path] = None,
    new_stem: ty.Optional[str] = None,
    copy_mode: FileSet.CopyMode = FileSet.CopyMode.copy,
) -> DicomImage:
    if out_dir is None:
        out_dir = Path(tempfile.mkdtemp())
    out_dir.mkdir(parents=True, exist_ok=True)
    dcm = dicom.load()
    dcm.PatientBirthDate = dcm.PatientBirthDate[:4] + "0101"
    dcm.PatientName = "Anonymous^Anonymous"
    for field in FIELDS_TO_DEIDENTIFY:
        try:
            elem = dcm[field]
        except KeyError:
            pass
        else:
            elem.value = ""
    return dicom.new(out_dir / dicom.fspath.name, dcm)


@extra_implementation(MedicalImage.deidentify)
def dicom_collection_deidentify(
    collection: DicomCollection,
    out_dir: ty.Optional[Path] = None,
    new_stem: ty.Optional[str] = None,
    copy_mode: FileSet.CopyMode = FileSet.CopyMode.copy,
) -> DicomCollection:
    if out_dir is None:
        out_dir = Path(tempfile.mkdtemp())
    if isinstance(collection, DicomDir):
        out_dir /= collection.name
    out_dir.mkdir(parents=True, exist_ok=True)
    deid_fspaths = []
    for dicom in collection.contents:
        deid_fspaths.append(dicom.deidentify(out_dir).fspath)
    type_ = type(collection)
    if isinstance(collection, DicomDir):
        deidentified = type_(out_dir)
    else:
        deidentified = type_(deid_fspaths)
    return deidentified


FIELDS_TO_DEIDENTIFY = [
    ("0008", "0014"),  # Instance Creator UID
    ("0008", "1111"),  # Referenced Performed Procedure Step SQ
    ("0008", "1120"),  # Referenced Patient SQ
    ("0008", "1140"),  # Referenced Image SQ
    ("0008", "0096"),  # Referring Physician Identification SQ
    ("0008", "1032"),  # Procedure Code SQ
    ("0008", "1048"),  # Physician(s) of Record
    ("0008", "1049"),  # Physician(s) of Record Identification SQ
    ("0008", "1050"),  # Performing Physicians' Name
    ("0008", "1052"),  # Performing Physician Identification SQ
    ("0008", "1060"),  # Name of Physician(s) Reading Study
    ("0008", "1062"),  # Physician(s) Reading Study Identification SQ
    ("0008", "1110"),  # Referenced Study SQ
    ("0008", "1111"),  # Referenced Performed Procedure Step SQ
    ("0008", "1250"),  # Related Series SQ
    ("0008", "9092"),  # Referenced Image Evidence SQ
    ("0008", "0080"),  # Institution Name
    ("0008", "0081"),  # Institution Address
    ("0008", "0082"),  # Institution Code Sequence
    ("0008", "0092"),  # Referring Physician's Address
    ("0008", "0094"),  # Referring Physician's Telephone Numbers
    ("0008", "009C"),  # Consulting Physician's Name
    ("0008", "1070"),  # Operators' Name
    ("0010", "4000"),  # Patient Comments
    # ("0010", "0010"),  # Patient's Name
    ("0010", "0021"),  # Issuer of Patient ID
    ("0010", "0032"),  # Patient's Birth Time
    ("0010", "0050"),  # Patient's Insurance Plan Code SQ
    ("0010", "0101"),  # Patient's Primary Language Code SQ
    ("0010", "1000"),  # Other Patient IDs
    ("0010", "1001"),  # Other Patient Names
    ("0010", "1002"),  # Other Patient IDs SQ
    ("0010", "1005"),  # Patient's Birth Name
    ("0010", "1010"),  # Patient's Age
    ("0010", "1040"),  # Patient's Address
    ("0010", "1060"),  # Patient's Mother's Birth Name
    ("0010", "1080"),  # Military Rank
    ("0010", "1081"),  # Branch of Service
    ("0010", "1090"),  # Medical Record Locator
    ("0010", "2000"),  # Medical Alerts
    ("0010", "2110"),  # Allergies
    ("0010", "2150"),  # Country of Residence
    ("0010", "2152"),  # Region of Residence
    ("0010", "2154"),  # Patient's Telephone Numbers
    ("0010", "2160"),  # Ethnic Group
    ("0010", "2180"),  # Occupation
    ("0010", "21A0"),  # Smoking Status
    ("0010", "21B0"),  # Additional Patient History
    ("0010", "21C0"),  # Pregnancy Status
    ("0010", "21D0"),  # Last Menstrual Date
    ("0010", "21F0"),  # Patient's Religious Preference
    ("0010", "2203"),  # Patient's Sex Neutered
    ("0010", "2297"),  # Responsible Person
    ("0010", "2298"),  # Responsible Person Role
    ("0010", "2299"),  # Responsible Organization
    ("0020", "9221"),  # Dimension Organization SQ
    ("0020", "9222"),  # Dimension Index SQ
    ("0038", "0010"),  # Admission ID
    ("0038", "0011"),  # Issuer of Admission ID
    ("0038", "0060"),  # Service Episode ID
    ("0038", "0061"),  # Issuer of Service Episode ID
    ("0038", "0062"),  # Service Episode Description
    ("0038", "0500"),  # Patient State
    ("0038", "0100"),  # Pertinent Documents SQ
    ("0040", "0260"),  # Performed Protocol Code SQ
    ("0088", "0130"),  # Storage Media File-Set ID
    ("0088", "0140"),  # Storage Media File-Set UID
    ("0400", "0561"),  # Original Attributes Sequence
    ("5200", "9229"),  # Shared Functional Groups SQ
]



import pytest
from fileformats.medimage import (
    NiftiGzX,
    NiftiGzXBvec,
    Analyze,
)
from logging import getLogger

from fileformats.medimage.dicom import DicomDir


logger = getLogger("fileformats")


def test_dicom_to_nifti(dummy_t1w_dicom: DicomDir) -> None:

    nifti_gz_x = NiftiGzX.convert(dummy_t1w_dicom)
    assert nifti_gz_x.metadata["json"]["EchoTime"] == 0.00207


def test_dicom_to_nifti_select_echo(dummy_magfmap_dicom: DicomDir) -> None:

    nifti_gz_x_e1 = NiftiGzX.convert(dummy_magfmap_dicom, file_postfix="_e1")
    nifti_gz_x_e2 = NiftiGzX.convert(dummy_magfmap_dicom, file_postfix="_e2")
    assert nifti_gz_x_e1.metadata["json"]["EchoNumber"] == 1
    assert nifti_gz_x_e2.metadata["json"]["EchoNumber"] == 2


def test_dicom_to_nifti_select_suffix(dummy_mixedfmap_dicom: DicomDir) -> None:

    nifti_gz_x_ph = NiftiGzX.convert(dummy_mixedfmap_dicom, file_postfix="_ph")
    nifti_gz_x_imaginary = NiftiGzX.convert(
        dummy_mixedfmap_dicom, file_postfix="_imaginary"
    )
    nifti_gz_x_real = NiftiGzX.convert(dummy_mixedfmap_dicom, file_postfix="_real")

    assert list(nifti_gz_x_ph.dims()) == [256, 256, 60]
    assert list(nifti_gz_x_imaginary.dims()) == [256, 256, 60]
    assert list(nifti_gz_x_real.dims()) == [256, 256, 60]


@pytest.mark.xfail
def test_dicom_to_nifti_with_extract_volume(dummy_dwi_dicom: DicomDir) -> None:

    nifti_gz_x_e1 = NiftiGzX.convert(dummy_dwi_dicom, extract_volume=30)
    assert nifti_gz_x_e1.metadata["dim"][0] == 3


def test_dicom_to_nifti_with_jq_edit(dummy_t1w_dicom: DicomDir) -> None:

    nifti_gz_x = NiftiGzX.convert(dummy_t1w_dicom, side_car_jq=".EchoTime *= 1000")
    assert nifti_gz_x.metadata["json"]["EchoTime"] == 2.07


def test_dicom_to_niftix_with_fslgrad(dummy_dwi_dicom: DicomDir) -> None:

    logger.debug("Performing FSL grad conversion")

    nifti_gz_x_fsgrad = NiftiGzXBvec.convert(dummy_dwi_dicom)

    bvec_mags = [
        (v[0] ** 2 + v[1] ** 2 + v[2] ** 2)
        for v in nifti_gz_x_fsgrad.encoding.directions
        if any(v)
    ]

    assert all(b in (0.0, 3000.0) for b in nifti_gz_x_fsgrad.encoding.b_values)
    assert len(bvec_mags) == 60
    assert all(abs(1 - m) < 1e5 for m in bvec_mags)


@pytest.mark.xfail
def test_dicom_to_nifti_as_4d(dummy_t1w_dicom: DicomDir) -> None:

    nifti_gz_x_e1 = NiftiGzX.convert(dummy_t1w_dicom, to_4d=True)
    assert nifti_gz_x_e1.metadata["dim"][0] == 4


@pytest.mark.xfail
def test_dicom_to_analyze(dummy_t1w_dicom: DicomDir) -> None:
    Analyze.convert(dummy_t1w_dicom)



import sys
import typing as ty
from pathlib import Path
import logging
from fileformats.core import extra, FileSet, mtime_cached_property
from fileformats.core.mixin import WithClassifiers
from .contents import ContentsClassifier
from .contents.imaging.modality import ImagingModality
from .contents.imaging.derivatives import Derivative
from .contents.anatomical_entity import AnatomicalEntity

logger = logging.getLogger("fileformats")

if sys.version_info >= (3, 9):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias
if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

if ty.TYPE_CHECKING:
    import numpy.typing  # noqa: F401


# =====================================================================
# Custom loader functions for different image types
# =====================================================================

DataArrayType: TypeAlias = (
    ty.Any
)  # In Py<3.9 this is problematic "numpy.typing.NDArray[typing.Union[numpy.floating[typing.Any], numpy.integer[typing.Any]]]"


class MedicalImagingData(FileSet):
    """Base class for all medical imaging data including pre-image raw data and
    associated data"""

    contains_phi: bool = True

    @extra
    def deidentify(
        self,
        out_dir: ty.Optional[Path] = None,
        new_stem: ty.Optional[str] = None,
        copy_mode: FileSet.CopyMode = FileSet.CopyMode.copy,
    ) -> Self:
        """Returns a new copy of the image with any subject-identifying information
        stripped from the from the image header"""
        raise NotImplementedError


class MedicalImage(WithClassifiers, MedicalImagingData):

    INCLUDE_HDR_KEYS: ty.Optional[ty.Tuple[str, ...]] = None
    IGNORE_HDR_KEYS: ty.Optional[ty.Tuple[str, ...]] = None
    binary = True
    classifiers_attr_name = "image_contents"
    image_contents = ()
    allowed_classifiers = (ContentsClassifier,)
    exclusive_classifiers = (ImagingModality, AnatomicalEntity, Derivative)

    @extra
    def read_array(self) -> DataArrayType:
        """
        Returns the binary data of the image in a numpy array
        """

    @mtime_cached_property
    def data_array(self) -> DataArrayType:
        return self.read_array()

    @extra
    def vox_sizes(self) -> ty.Tuple[float, float, float]:
        """The length of the voxels along each dimension"""
        raise NotImplementedError

    @extra
    def dims(self) -> ty.Tuple[int, int, int]:
        """The dimensions of the image"""
        raise NotImplementedError



import sys
import os
import typing as ty
from collections import defaultdict, Counter
from pathlib import Path
from fileformats.core.decorators import mtime_cached_property
from fileformats.core import extra, FileSet, extra_implementation
from fileformats.core.collection import TypedCollection
from fileformats.generic import TypedDirectory, TypedSet
from fileformats.application import Dicom
from .base import MedicalImage

if sys.version_info >= (3, 9):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias
if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

if ty.TYPE_CHECKING:
    import pydicom.tag

    TagListType: TypeAlias = ty.Union[
        ty.List[int],
        ty.List[str],
        ty.List[ty.Tuple[int, int]],
        ty.List[pydicom.tag.BaseTag],
    ]
# =====================================================================
# Custom loader functions for different image types
# =====================================================================


class DicomImage(MedicalImage, Dicom):
    """A DICOM file that contains image data. Derives from the generic Dicom class in
    the `application` namespace as well as medical image base class"""


def dicom_sort_key(dicom: Dicom) -> str:
    """Sorts DICOM objects by SOPInstanceUID"""
    assert isinstance(dicom.metadata, dict)
    return dicom.metadata["SOPInstanceUID"]  # type: ignore[no-any-return]


class DicomCollection(MedicalImage, TypedCollection):
    """Base class for collections of DICOM files, which can either be stored within a
    directory (DicomDir) or presented as a flat list (DicomSeries)
    """

    content_types = (DicomImage,)

    def __len__(self) -> int:
        return len(self.contents)

    @extra
    def series_number(self) -> str:
        raise NotImplementedError


class DicomDir(TypedDirectory, DicomCollection):
    content_types = (DicomImage,)

    @mtime_cached_property
    def contents(self) -> ty.List[DicomImage]:
        return sorted(TypedDirectory.contents.__get__(self), key=dicom_sort_key)


class DicomSeries(TypedSet, DicomCollection):
    content_types = (DicomImage,)

    @classmethod
    def from_paths(
        cls,
        fspaths: ty.Iterable[Path],
        common_ok: bool = False,
        **kwargs: ty.Any,
    ) -> ty.Tuple[ty.Set[Self], ty.Set[Path]]:
        """Separates a list of DICOM files into separate series from the file-system
        paths

        Parameters
        ----------
        fspaths : ty.Iterable[Path]
            the fspaths pointing to the DICOM files
        common_ok : bool, optional
            included to match the signature of the overridden method, but ignored as each
            dicom should belong to only one series.
        specific_tags : ty.Optional[TagListType], optional
            the DICOM tags to read from the files. If None, the default tags will be
            read
        **kwargs : ty.Any
            additional keyword arguments to passed through to the DicomImage constructor

        Returns
        -------
        tuple[set[DicomSeries], set[Path]]
            the found dicom series objects and any unrecognised file paths
        """
        dicoms, remaining = DicomImage.from_paths(
            fspaths, common_ok=common_ok, **kwargs
        )
        series_dict = defaultdict(list)
        for dicom in dicoms:
            metadata = dicom.read_metadata(metadata_keys=cls.ID_KEYS)
            series_dict[tuple(metadata[k] for k in cls.ID_KEYS)].append(dicom)
        return set([cls(d.fspath for d in s) for s in series_dict.values()]), remaining

    @mtime_cached_property
    def contents(self) -> ty.List[DicomImage]:
        return sorted(TypedSet.contents.__get__(self), key=dicom_sort_key)

    ID_KEYS = ("StudyInstanceUID", "SeriesNumber")


@extra_implementation(FileSet.read_metadata)
def dicom_collection_read_metadata(
    collection: DicomCollection, **kwargs: ty.Any
) -> ty.Mapping[str, ty.Any]:
    # Collated DICOM headers across series
    collated: ty.Dict[str, ty.Any] = {}
    key_repeats: ty.Counter[str] = Counter()
    varying_keys = set()
    # We use the "contents" property implementation in TypeSet instead of the overload
    # in DicomCollection because we don't want the metadata to be read ahead of the
    # the `select_metadata` call below
    base_class: ty.Union[ty.Type[TypedSet], ty.Type[TypedDirectory]] = (
        TypedSet if isinstance(collection, DicomSeries) else TypedDirectory
    )
    for dicom in base_class.contents.__get__(collection):
        for key, val in dicom.metadata.items():
            try:
                prev_val = collated[key]
            except KeyError:
                collated[
                    key
                ] = val  # Insert initial value (should only happen on first iter)
                key_repeats.update([key])
            else:
                if key in varying_keys:
                    collated[key].append(val)
                # Check whether the value is the same as the values in the previous
                # images in the series
                elif val != prev_val:
                    collated[key] = [prev_val] * key_repeats[key] + [val]
                    varying_keys.add(key)
                else:
                    key_repeats.update([key])
    return collated


def get_dicom_tag(
    file: ty.Union[str, os.PathLike[ty.Any], ty.BinaryIO],
    target_tag: ty.Tuple[int, int],
) -> ty.Union[str, bytes, None]:
    """A basic function to read a DICOM file and extract the value of a specific tag.
    This is a low-level function that does not use any external libraries.
    It is not a replacement for pydicom, but can be used to extract specific tags
    without loading the entire DICOM file into memory.

    Parameters
    ----------
    filepath : str or os.PathLike
        The path to the DICOM file.
    target_tag : tuple[int, int]
        The DICOM tag to extract, specified as a tuple of (group, element).
        For example, (0x0010, 0x0010) for PatientName.

    Returns
    -------
    str or bytes or None
        The value of the specified DICOM tag, decoded as a string if possible.
        If the tag is not found or cannot be decoded, returns None.
    """
    if isinstance(file, (str, os.PathLike)):
        filepath = file
        file_stream = open(filepath, "rb")
        close_stream = True
    elif hasattr(file, "read"):
        file_stream = file  # type: ignore[assignment]
        close_stream = False
    else:
        raise TypeError("file must be a path-like object or a binary stream")

    try:
        file_stream.seek(132)  # Skip preamble and 'DICM' if at file start

        while True:
            tag_bytes = file_stream.read(4)
            if len(tag_bytes) < 4:
                break

            group = int.from_bytes(tag_bytes[:2], "little")
            element = int.from_bytes(tag_bytes[2:], "little")
            tag = (group, element)

            vr_bytes = file_stream.read(2)
            # If the VR bytes are not purely alphabetic, assume implicit VR.
            if vr_bytes.decode(errors="ignore").strip().isalpha():
                vr = vr_bytes.decode()
                if vr in {"OB", "OW", "OF", "SQ", "UT", "UN"}:
                    file_stream.read(2)  # reserved
                    length = int.from_bytes(file_stream.read(4), "little")
                else:
                    length = int.from_bytes(file_stream.read(2), "little")
            else:
                # Implicit VR: rewind the 2 bytes and read a 4-byte length
                file_stream.seek(-2, 1)
                length = int.from_bytes(file_stream.read(4), "little")
            value = file_stream.read(length)

            if tag == target_tag:
                try:
                    return value.decode().strip()
                except UnicodeDecodeError:
                    return value
    finally:
        if close_stream:
            file_stream.close()

    return None  # Not found


# class Vnd_Siemens_Vision(DicomImage):
#     ext = ".ima"


# class Vnd_Siemens_VisionDir(DicomDir):
#     content_types = (Vnd_Siemens_Vision,)



from pathlib import Path
import json
import typing as ty
import tempfile
from fileformats.core import converter
from pydra.compose import python, workflow
from fileformats.medimage.dicom import DicomDir, DicomCollection, DicomSeries
from fileformats.application import Json
from fileformats.medimage import (
    MedicalImage,
    Analyze,
    Nifti,
    NiftiGz,
    NiftiX,
    NiftiGzX,
    NiftiXBvec,
    NiftiBvec,
    NiftiGzBvec,
    NiftiGzXBvec,
)
from fileformats.core.typing import PathType
from pydra.tasks.dcm2niix import Dcm2Niix
from pydra.tasks.mrtrix3.v3_1 import MrConvert
from fileformats.generic import File, Directory  # noqa: F401
from fileformats.vendor.mrtrix3.medimage import ImageIn, ImageOut, Tracks  # noqa: F401


@python.define  # type: ignore[untyped-decorator]
def EnsureDicomDir(dicom: DicomCollection) -> DicomDir:
    if isinstance(dicom, DicomSeries):
        dicom_dir_fspath = tempfile.mkdtemp()
        dicom.copy(dicom_dir_fspath, mode=DicomDir.CopyMode.link)
        dicom = DicomDir(dicom_dir_fspath)
    elif not isinstance(dicom, DicomDir):
        raise RuntimeError(
            "Unrecognised input to ensure_dicom_dir, should be DicomSeries or DicomDir "
            f"not {dicom}"
        )
    return dicom


@converter(source_format=DicomCollection, target_format=Nifti)  # type: ignore[untyped-decorator]
@converter(source_format=DicomCollection, target_format=NiftiGz, compress="y")  # type: ignore[untyped-decorator]
@converter(source_format=DicomCollection, target_format=NiftiX)  # type: ignore[untyped-decorator]
@converter(source_format=DicomCollection, target_format=NiftiGzX, compress="y")  # type: ignore[untyped-decorator]
@converter(source_format=DicomCollection, target_format=NiftiXBvec)  # type: ignore[untyped-decorator]
@converter(source_format=DicomCollection, target_format=NiftiBvec)  # type: ignore[untyped-decorator]
@converter(source_format=DicomCollection, target_format=NiftiGzBvec)  # type: ignore[untyped-decorator]
@converter(source_format=DicomCollection, target_format=NiftiGzXBvec, compress="y")  # type: ignore[untyped-decorator]
@workflow.define(outputs=["out_file"])  # type: ignore[untyped-decorator]
def ExtendedDcm2niix(
    in_file: DicomCollection,
    compress: str = "n",
    file_postfix: ty.Optional[str] = None,
    side_car_jq: ty.Optional[str] = None,
    extract_volume: ty.Optional[int] = None,
    to_4d: bool = False,
) -> Nifti:
    """The Dcm2niix command wrapped in a workflow in order to map the inputs and outputs
    onto "in_file" and "out_file", respectively, and implement optional post-conversion
    manipulations to allow manual override of conversion issues.

    Parameters
    ----------
    name : str
        name of the workflow
    compress : str, optional
        whether to apply compression to the conversion, by default "n"
    file_postfix : str, optional
        select one of the multiple output files by its generated postfix.
        See https://github.com/rordenlab/dcm2niix/blob/master/FILENAMING.md for
        complete list of different postfixes that will be generated, by default None
    side_car_jq : str, optional
        JQ (https://stedolan.github.io/jq/) expression that can be provided to edit
        the resulting JSON side-car. Can be used to fix any conversion issues or add
        required fields manually, by default None
    extract_volume : int, optional
        extract a 3D volume from a 4D dataset by passing the index (0-based) of the
        volume to extract, by default None
    to_4d : bool, optional
        whether to wrap resulting 3D NIfTI volume in a 4D dataset, by default False

    Returns
    -------
    out_file: list[Path]
        the converter workflow

    Raises
    ------
    ValueError
        when mutually exclusive "extract_volume" and "to_4d" options are provided
    """

    if extract_volume is not None and to_4d:
        raise ValueError(
            f"'extract_volume' ({extract_volume}) and 'to_4d' are mutually exclusive"
        )
    # Create workflow to map input field to "in_file" and optionally perform post-conversion
    # steps to manipulate the converted NIfTI files
    ensure_dicom_dir = workflow.add(EnsureDicomDir(dicom=in_file))

    dcm2niix = workflow.add(
        Dcm2Niix(
            in_dir=ensure_dicom_dir.out,
            out_dir=Path("."),
            compress=compress,
            file_postfix=file_postfix,
        )
    )
    out_file = dcm2niix.out_file
    out_json = dcm2niix.out_json
    # Add MRConvert step to either select a single volume of a 4D dataset or the inverse,
    # wrap a single volume in a 4D dataset
    if extract_volume is not None or to_4d:
        if extract_volume:
            coord = (3, extract_volume)
            axes = [0, 1, 2]
        else:  # to_4d
            coord = None
            axes = [0, 1, 2, -1]
        mrconvert = workflow.add(
            MrConvert(
                in_file=out_file,
                out_file="out" + NiftiGzX.ext,
                coord=coord,
                axes=axes,
            )
        )
        out_file = mrconvert.out_file
    # Add JQ edit of side car to allow manual fixing of any conversion issues
    if side_car_jq is not None:
        json_edit = workflow.add(
            EditDcm2niixSideCar(in_file=out_json, jq_expr=side_car_jq), name="json_edit"
        )
        out_json = json_edit.out

    collect_outputs = workflow.add(
        CollectDcm2niixOutputs(
            out_file=out_file,
            out_json=out_json,
            out_bvec=dcm2niix.out_bvec,
            out_bval=dcm2niix.out_bval,
        )
    )
    # Set workflow outputs
    return collect_outputs.out  # type: ignore[no-any-return]


@python.define  # type: ignore[untyped-decorator]
def EditDcm2niixSideCar(
    in_file: ty.Optional[Json], jq_expr: str, out_file: ty.Optional[PathType] = None
) -> ty.Optional[Json]:
    """ "Applies ad-hoc edit of JSON side car with JQ query language"""
    if in_file is None:
        assert jq_expr is not None
        return None
    if out_file is None:
        out_dir = Path(tempfile.mkdtemp())
        out_file = out_dir / "out.json"
    with open(in_file) as f:
        dct = json.load(f)
    try:
        import jq
    except ImportError:
        raise RuntimeError(
            "Cannot edit dcm2niix file with jq expression as `jq` package is not "
            "available in the current environment. Please reinstall "
            "fileformats-medimage-extras with the `jq` extra, i.e. "
            "`pip install 'fileformats-medimage-extras[jq]'`, noting that on Windows "
            "you will first need to install the jq package from https://jqlang.org/download/."
        )
    dct = jq.compile(jq_expr).input(dct).first()
    with open(out_file, "w") as f:
        json.dump(dct, f)
    return Json(out_file)


@python.define  # type: ignore[untyped-decorator]
def CollectDcm2niixOutputs(
    out_file: Path,
    out_json: ty.Optional[Path],
    out_bvec: ty.Optional[Path],
    out_bval: ty.Optional[Path],
) -> Nifti:
    klass = NiftiGz if out_file.suffix == ".gz" else Nifti
    if out_json:
        if klass is NiftiGz:
            klass = NiftiGzX
        else:
            klass = NiftiX
    if any((out_bvec, out_bval)):
        if not all((out_bvec, out_bval)):
            raise RuntimeError(
                "Both bvec and bval files must be present for diffusion NIfTI outputs"
            )
        if klass is NiftiGz:
            klass = NiftiGzXBvec
        elif klass is NiftiGzX:
            klass = NiftiGzXBvec
        elif klass is NiftiX:
            klass = NiftiXBvec
        else:
            assert klass is Nifti
            klass = NiftiBvec

    return klass(
        [out_file] + [f for f in (out_json, out_bvec, out_bval) if f is not None]
    )


converter(
    source_format=MedicalImage,
    target_format=Analyze,
    out_file_="out_file" + Analyze.ext,
)(MrConvert)



from fileformats.application import Gzip
from fileformats.generic import BinaryFile
from fileformats.core import validated_property
from fileformats.core.mixin import WithSeparateHeader, WithMagicVersion
from .base import MedicalImage
from fileformats.core.exceptions import FormatMismatchError


# ==================
# Other Data Formats
# ==================


class AnalyzeHeader(BinaryFile):

    ext = ".hdr"
    contains_phi = False


class Analyze(WithSeparateHeader, MedicalImage, BinaryFile):

    ext = ".img"
    header_type = AnalyzeHeader


class Mgh(WithMagicVersion, BinaryFile):
    """
    FreeSurfer 4-dimensional brain images

    See https://surfer.nmr.mgh.harvard.edu/fswiki/FsTutorial/MghFormat
    """

    ext = ".mgh"
    magic_pattern = rb"(....)"  # First integer is the version string

    @validated_property
    def _is_supported_version(self) -> None:
        assert isinstance(self.version, str)
        if int(self.version) != 1:
            raise FormatMismatchError(
                f"Unsupported version {self.version} found in MGH format {self}"
            )


class MghGz(Gzip[Mgh]):  # type: ignore[type-arg]
    """
    FreeSurfer 4-dimensional brain images, gzipped

    See https://surfer.nmr.mgh.harvard.edu/fswiki/FsTutorial/MghFormat
    """

    iana_mime = "application/x-mgh+zip"
    ext = ".mgz"



import typing as ty
from fileformats.core.mixin import WithMagicNumber
from fileformats.image import (
    RasterImage,
    VectorImage,
    Bitmap,
    Jpeg,
    Tiff,
)
from fileformats.application import Gzip
from .nifti import Nifti1, NiftiGz
from .dicom import DicomImage


class GDCM(DicomImage):
    # iana_mime = None
    pass


class GIPL(RasterImage):
    ext = ".gipl"


class VTK(VectorImage):
    ext = ".vtk"


class PGM(RasterImage):
    ext = ".pgm"


class MetaImage(RasterImage):
    ext = ".mhd"


class Nrrd(WithMagicNumber, RasterImage):

    ext = ".nrrd"
    alternate_exts = (".nhdr",)
    magic_number = b"NRRD"


class NrrdGz(Gzip[Nrrd]):  # type: ignore[type-arg]

    ext = ".nrrd.gz"
    alternate_exts = (".nhdr.gz",)


ItkImage = ty.Union[
    Nifti1, NiftiGz, DicomImage, Bitmap, Tiff, Jpeg, GIPL, MetaImage, Nrrd, NrrdGz, PGM
]

ItkAll = ty.Union[
    Nifti1,
    NiftiGz,
    DicomImage,
    Bitmap,
    Tiff,
    Jpeg,
    GIPL,
    MetaImage,
    Nrrd,
    NrrdGz,
    PGM,
    VTK,
]

__all__ = [
    "GDCM",
    "GIPL",
    "VTK",
    "PGM",
    "MetaImage",
    "Nrrd",
    "NrrdGz",
    "ItkImage",
    "ItkAll",
]



import numpy as np
import typing as ty
from pathlib import Path
from fileformats.core import FileSet, extra_implementation
from fileformats.core.sampling import SampleFileGenerator
from fileformats.medimage import DwiEncoding, Bval, Bvec
from fileformats.medimage.diffusion import EncodingArrayType


@extra_implementation(Bval.read_array)
def bval_read_array(bval: Bval) -> EncodingArrayType:
    return np.asarray([float(ln) for ln in bval.read_contents().split()])


@extra_implementation(DwiEncoding.read_encodings)
def bvec_read_array(bvec: Bvec) -> EncodingArrayType:
    bvals = bvec.b_values_file.read_array()
    directions = np.asarray(
        [[float(x) for x in ln.split()] for ln in bvec.read_contents().splitlines()]
    ).T
    return np.concatenate((directions, bvals.reshape((-1, 1))), axis=1)


@extra_implementation(FileSet.generate_sample_data)
def bval_generate_sample_data(
    bval: Bval, generator: SampleFileGenerator
) -> ty.List[Path]:
    bvals_fspath = generator.generate_fspath(Bval)
    bvals = [
        str(generator.rng.randrange(0, 5000))
        for _ in range(generator.rng.randrange(5, 100))
    ]
    with open(bvals_fspath, "w") as f:
        f.write(" ".join(bvals))
    return [bvals_fspath]


@extra_implementation(FileSet.generate_sample_data)
def bvec_generate_sample_data(
    bvec: Bvec, generator: SampleFileGenerator
) -> ty.List[Path]:
    bvecs_fspath = generator.generate_fspath(Bvec)
    bvecs = np.asarray(
        [
            [generator.rng.uniform(0, 1) for _ in range(3)]
            for _ in range(generator.rng.randrange(5, 100))
        ]
    ).T
    # Normalise bvecs
    bvecs = bvecs / np.sqrt(bvecs[0, :] ** 2 + bvecs[1, :] ** 2 + bvecs[2, :] ** 2)
    np.savetxt(bvecs_fspath, bvecs)
    bvals = [str(generator.rng.randrange(0, 5000)) for _ in range(bvecs.shape[1])]
    bvals_fspath = bvecs_fspath.parent / (bvecs_fspath.stem + ".bval")
    with open(bvals_fspath, "w") as f:
        f.write(" ".join(bvals))
    return [bvecs_fspath, bvals_fspath]
